/*
 * calmwm - the calm window manager
 *
 * Copyright (c) 2004 Marius Aamodt Eriksen <marius@monkey.org>
 * All rights reserved.
 *
 * $Id: search.c,v 1.8 2004/07/10 20:48:13 marius Exp $
 */

#include "headers.h"
#include "calmwm.h"

#define SEARCH_MAXLEN 256

#define SearchMask (KeyPressMask|ExposureMask)

static int  _strsubmatch(char *, char *);
static void _matchq(struct client_ctx_q *, char *);

void
search_init(struct screen_ctx *sc)
{
	sc->searchwin = XCreateSimpleWindow(G_dpy, sc->rootwin, 0, 0,
	    1, 1, 1, sc->blackpix, sc->whitepix);
}

void
search_start(void)
{
	struct screen_ctx *sc = screen_current();
	int x, y, dx, dy, fontheight, focusrevert, mutated;
	XEvent e;
	char searchstr[SEARCH_MAXLEN];
	char dispstr[SEARCH_MAXLEN + sizeof("search>") - 1];
	Window focuswin;
	struct client_ctx *cc = NULL, *old_cc = client_current();
	char chr;
	enum ctltype ctl;
	size_t len;
	struct client_ctx_q searchq;
	u_int n;

	TAILQ_INIT(&searchq);

	xu_ptr_getpos(sc->rootwin, &x, &y);

	searchstr[0] = '\0';

	dy = fontheight = G_font->ascent + G_font->descent + 1;
	dx = XTextWidth(G_font, "search>", 7);

	XMoveResizeWindow(G_dpy, sc->searchwin, x, y, dx, dy);
	XSelectInput(G_dpy, sc->searchwin, SearchMask);
	XMapRaised(G_dpy, sc->searchwin);

	XGetInputFocus(G_dpy, &focuswin, &focusrevert);
	XSetInputFocus(G_dpy, sc->searchwin,
	    RevertToPointerRoot, CurrentTime);

	for (;;) {
		XMaskEvent(G_dpy, SearchMask, &e);

		switch (e.type) {
		case KeyPress:
			if (input_keycodetrans(e.xkey.keycode, &ctl, &chr, 1) < 0)
				continue;

			mutated = 0;

			switch (ctl) {
			case CTL_ERASEONE:
				if ((len = strlen(searchstr)) > 0)
					searchstr[len - 1] = '\0';
				mutated = 1;
				break;
			case CTL_UP:
				cc = TAILQ_LAST(&searchq, client_ctx_q);
				if (cc == NULL)
					break;

				TAILQ_REMOVE(&searchq, cc, searchentry);
				TAILQ_INSERT_HEAD(&searchq, cc, searchentry);
				break;
			case CTL_DOWN:
				cc = TAILQ_FIRST(&searchq);
				if (cc == NULL)
					break;

				TAILQ_REMOVE(&searchq, cc, searchentry);
				TAILQ_INSERT_TAIL(&searchq, cc, searchentry);
				break;
			case CTL_RETURN:
				/* This is just picking the match the
				 * cursor is over. */
				if (strlen(searchstr) == 0)
					goto out;

				if ((cc = TAILQ_FIRST(&searchq)) != NULL)
					goto found;
				else
					goto out;
			case CTL_ABORT:
				goto out;
			default:
				break;
			}

			if (chr != '\0') {
				char str[2];

				str[0] = chr;
				str[1] = '\0';
				mutated = 1;
				strlcat(searchstr, str, sizeof(searchstr));
			}

			if (mutated && strlen(searchstr) > 0)
				_matchq(&searchq, searchstr);
			else if (mutated)
				TAILQ_INIT(&searchq);

		case Expose:
			snprintf(dispstr, sizeof(dispstr), "search>%s", searchstr);
			dx = XTextWidth(G_font, dispstr, strlen(dispstr));
			dy = fontheight;

			TAILQ_FOREACH(cc, &searchq, searchentry) {
				dx = MAX(dx,
				    XTextWidth(G_font, cc->name,
					strlen(cc->name)));
				dy += fontheight;
			}

			XClearWindow(G_dpy, sc->searchwin);
			XResizeWindow(G_dpy, sc->searchwin, dx, dy);

			XDrawString(G_dpy, sc->searchwin, sc->gc, 0,
			    G_font->ascent + 1,
			    dispstr, strlen(dispstr));
			n = 1;
			TAILQ_FOREACH(cc, &searchq, searchentry) {
				XDrawString(G_dpy, sc->searchwin, sc->gc, 0,
				    n*fontheight + G_font->ascent + 1,
				    cc->name, strlen(cc->name));
				n++;
			}

			if (n > 1)
				XFillRectangle(G_dpy, sc->searchwin, sc->gc,
				    0, fontheight, dx, fontheight);

			break;
		}
	}

 out:
	/* (if no match) */
	XSetInputFocus(G_dpy, focuswin,
	    focusrevert, CurrentTime);
 found:
	XUnmapWindow(G_dpy, sc->searchwin);

	if (cc != NULL) {
		if (cc->flags & CLIENT_HIDDEN)
			client_unhide(cc);

		cycle_current(sc->cycle, cc);

		if (old_cc)
			client_ptrsave(old_cc);
		client_ptrwarp(cc);
	}
}

static void
_matchq(struct client_ctx_q *matchq, char *search)
{
	struct client_ctx *cc;
	struct client_ctx_q tier2;
	struct winname *wn;

	TAILQ_INIT(matchq);
	TAILQ_INIT(&tier2);

	/*
	 * 1. Look through labels.
	 * 
	 * 2. Look at title history, from present to past.
	 */

	TAILQ_FOREACH(cc, &G_clientq, entry) {
		if (cc->label != NULL && _strsubmatch(search, cc->label)) {
			cc->matchname = cc->label;
			TAILQ_INSERT_TAIL(matchq, cc, searchentry);
		} else {
			TAILQ_FOREACH_REVERSE(wn, &cc->nameq, winname_q, entry)
				if (_strsubmatch(search, wn->name)) {
					cc->matchname = wn->name;
					TAILQ_INSERT_TAIL(&tier2,
					    cc, searchentry);
					break;
				}
		}
	}

	while ((cc = TAILQ_FIRST(&tier2)) != NULL) {
		TAILQ_REMOVE(&tier2, cc, searchentry);
		TAILQ_INSERT_TAIL(matchq, cc, searchentry);
	}
}

static int
_strsubmatch(char *sub, char *str)
{
	size_t len, sublen;
	u_int n;

	len = strlen(str);
	sublen = strlen(sub);

	if (sublen > len)
		return (0);

	for (n = 0; n <= len - sublen; n++)
		if (strncasecmp(sub, str + n, sublen) == 0)
			return (1);

	return (0);
}
